@@ -5,7 +5,7 @@
 # file included as part of this package.
 #
 
-from nessaid_cli.tokens import CliToken, MATCH_FAILURE
+from nessaid_cli.tokens import CliToken, MATCH_FAILURE, NullTokenValue
 from nessaid_cli.lex_yacc_common import DollarVariable
 from nessaid_cli.utils import ExtendedString
 
@@ -13,6 +13,8 @@ from nessaid_cli.utils import ExtendedString
 class CliParameter(ExtendedString):
 
     def __init__(self, value, defvalue=None, has_def_value=False):
+        self.defvalue = None
+        self.has_def_value = False
         if type(has_def_value) is not bool:
             raise AttributeError("Attribute has_def_value should be Boolean")
         super().__init__(value, defvalue=defvalue, has_def_value=has_def_value)
@@ -61,263 +63,6 @@ class CliTokenArgument(CliArgument):
         self._param_name = param_name
 
 
-class TokenHierarchyElement():
-
-    def __init__(self, element, path):
-        self._element = element
-        self._path = tuple(path)
-        self._named_vars = {}
-        self._numbered_vars = {}
-        self._input_sequence = []
-
-    @property
-    def named_variables(self):
-        return self._named_vars
-
-    @property
-    def token_variables(self):
-        return self._numbered_vars
-
-    @property
-    def input_sequence(self):
-        return self._input_sequence
-
-    @property
-    def element(self):
-        return self._element
-
-    def add_named_variable(self, var):
-        self._named_vars[var.var_id] = var
-
-    def add_numbered_variable(self, var):
-        self._numbered_vars[var.var_id] = var
-
-    @property
-    def path(self):
-        return self._path
-
-    def __eq__(self, rhs):
-        if isinstance(rhs, TokenHierarchyElement):
-            return (self._element == rhs._element) and (self._path == rhs.path)
-        return False
-
-    def __hash__(self):
-        return hash(self._element) + hash(self._path)
-
-
-class TokenLookup():
-
-    def __init__(self, token, root_grammar=None):
-        self.token = token
-        self.path = []
-        self._root_grammar = root_grammar
-        self._parent_stack = []
-
-    @property
-    def name(self):
-        return self.token.value
-
-    def add_prefix(self, prefix):
-        self.path = [prefix] + self.path
-
-    def get_element_hierarchy(self):
-        self._parent_stack = []
-        hierarchy = []
-        level = 0
-        parent_path = self.path.copy()
-        while True:
-            parent = self.get_parent(level)
-            hierarchy.append(TokenHierarchyElement(parent, parent_path.copy()))
-            parent_path.pop()
-            level += 1
-            if parent is self._root_grammar:
-                self._parent_stack = []
-                return hierarchy
-
-    def __eq__(self, rhs):
-        if isinstance(rhs, TokenLookup):
-            if self._root_grammar is rhs._root_grammar:
-                if (self.token == rhs.token) and (self.path == rhs.path):
-                    return True
-        return False
-
-    def __hash__(self):
-        return hash(self.token) + hash(tuple(self.path))
-
-    def __repr__(self):
-        return "{}:<{}>".format(self.token, ",".join(str(p) for p in self.path))
-
-    def _locate(self, path):
-        parent = self._root_grammar
-
-        while path:
-            position = path.pop(0)
-            parent = parent.get(position)
-        return parent
-
-    def get_parent(self, parent_level):
-        """
-        R  P2  P1 T
-        0  p2  p1 pt
-        """
-        if not self._parent_stack:
-            level = 0
-            parent = self.token
-
-            self._parent_stack.append(parent)
-            level += 1
-
-            while parent:
-                if type(parent) in [NamedGrammar]:
-                    parent = self._locate(self.path[1:][:-level])
-                else:
-                    parent = parent.parent
-                self._parent_stack.append(parent)
-                if parent is self._root_grammar:
-                    break
-                level += 1
-
-        return self._parent_stack[parent_level]
-
-
-    def next(self):
-        nexts = []
-        self._parent_stack = []
-        # import pdb; pdb.set_trace()
-        level = 0
-        parent = self.get_parent(level + 1)
-        level += 1
-        position = self.token.position + 1
-        end_of_grammar = True
-
-        while parent:
-            elem = None
-            if not isinstance(parent, AlternativeInputElement):
-                while position < len(parent):
-                    elem = parent.get(position)
-                    if elem.repeat_count > 1:
-                        firsts = elem.get(0).first(root_grammar=self._root_grammar)
-                        for f in firsts:
-                            f.add_prefix(elem.position)
-                    else:
-                        firsts = elem.first(root_grammar=self._root_grammar)
-
-                    temp_parent = parent
-                    temp_level = level
-
-                    while temp_parent:
-                        for f in firsts:
-                            f.add_prefix(temp_parent.position)
-                        if temp_parent is self._root_grammar:
-                            break
-                        temp_parent = self.get_parent(temp_level + 1)
-                        temp_level += 1
-
-                    nexts += firsts
-                    if elem.mandatory:
-                        end_of_grammar = False
-                        break
-                    position += 1
-
-                    if parent.repeat_count > 1:
-                        break
-                if elem and elem.mandatory:
-                    break
-
-            if parent is self._root_grammar:
-                break
-            position = parent.position + 1
-            parent = self.get_parent(level + 1)
-            level += 1
-
-        if end_of_grammar or not nexts:
-            nexts.append(EndOfInpuToken())
-
-        self._parent_stack = []
-        return nexts
-
-
-class _EndOfInpuToken(TokenLookup, CliToken):
-
-    __instance = None
-
-    @property
-    def completable(self):
-        return False
-
-    def get_value(self, match_string=None):
-        return None
-
-    @staticmethod
-    def getInstance():
-        if _EndOfInpuToken.__instance == None:
-            _EndOfInpuToken()
-        return _EndOfInpuToken.__instance
-
-    def __init__(self):
-        if _EndOfInpuToken.__instance != None:
-            raise Exception("This class is a singleton!")
-        self._name = "$END_OF_INPUT"
-        _EndOfInpuToken.__instance = self
-
-    def __repr__(self):
-        return "< End of Input >"
-
-    def __str__(self):
-        return self.__repr__()
-
-    def __eq__(self, rhs):
-        if isinstance(rhs, _EndOfInpuToken):
-            return True
-        return False
-
-    def __hash__(self):
-        return hash(self._name)
-
-    def __not__(self):
-        return True
-
-    def truth(self):
-        return False
-
-    def is_(self, rhs):
-        if isinstance(rhs, _EndOfInpuToken):
-            return True
-        return False
-
-    def is_not(self, rhs):
-        if isinstance(rhs, _EndOfInpuToken):
-            return False
-        return True
-
-    def __nonzero__(self):
-        return False
-
-    def __bool__(self):
-        return False
-
-    def complete(self, *args, **kwargs):
-        return []
-
-    def match(self, token_input):
-        return MATCH_FAILURE
-
-    @property
-    def helpstring(self):
-        return str(self)
-
-    @property
-    def name(self):
-        return self._name
-
-    def next(self):
-        return [self]
-
-
-def EndOfInpuToken():
-    return _EndOfInpuToken.getInstance()
-
-
 class DuplicateDefinitionException(Exception):
 
     def __init__(self, name):
@@ -348,13 +93,13 @@ class ArgumentError(Exception):
 
 class InputElement():
 
-    def __init__(self):
+    def __init__(self, helpstring=None):
         self._parent = None
         self._value = None
-        self._pre_match_binding = []
-        self._post_match_binding = []
-        self._position = None
+        self._pre_match_binding = ()
+        self._post_match_binding = ()
         self._repeat_count = 1
+        self._help_string = helpstring
 
     def copy_extras(self, cp):
         cp._pre_match_binding = self._pre_match_binding
@@ -395,24 +140,8 @@ class InputElement():
             raise ValueError("Expected InputElementCollection object for parent")
         self._parent = p
 
-    @property
-    def position(self):
-        if self._position is None:
-            if type(self.parent) in [NamedGrammar, GrammarRefElement]:
-                self._position = 0
-            else:
-                self._position = self.parent.index(self)
-        return self._position
-
-    @position.setter
-    def position(self, pos):
-        pos = int(pos)
-        if pos < 0:
-            raise ValueError("Position should be 0 or a positive integer")
-        self._position = pos
-
     def validate_binding(self, b):
-        if not isinstance(b, list):
+        if not isinstance(b, tuple):
             for e in b:
                 if not isinstance(e, str):
                     raise ValueError("Expected str object for binding code")
@@ -439,62 +168,51 @@ class InputElement():
     def mandatory(self):
         return False
 
+    @property
+    def helpstring(self):
+        return self._help_string
+
 
 class ConstantInputElement(InputElement):
 
-    def __init__(self, keyword):
-        super().__init__()
+    def __init__(self, keyword, helpstring=None):
+        super().__init__(helpstring)
         self._value = keyword
 
     def copy(self):
-        cp = ConstantInputElement(self._value)
+        cp = ConstantInputElement(self._value, helpstring=self.helpstring)
         self.copy_extras(cp)
         return cp
 
-    def __repr__(self, verbose=True):
+    def __repr__(self, verbose=True): # noqa
         return self._value
 
     @property
     def mandatory(self):
         return True
 
-    def first(self, root_grammar=None):
-        first =  TokenLookup(self, root_grammar=root_grammar)
-        if self.parent:
-            first.add_prefix(self.position)
-        return [first]
-
 
 class KeywordInputElement(InputElement):
 
-    def __init__(self, keyword):
-        super().__init__()
+    def __init__(self, keyword, helpstring=None):
+        super().__init__(helpstring)
         if isinstance(keyword, str):
             self._value = keyword
         else:
             raise ValueError("Expected str object")
 
     def copy(self):
-        cp = KeywordInputElement(self._value)
+        cp = KeywordInputElement(self._value, helpstring=self.helpstring)
         self.copy_extras(cp)
         return cp
 
-    def __repr__(self, verbose=True):
+    def __repr__(self, verbose=True): # noqa
         return self._value
 
-    def __hash__(self):
-        return self._value.__hash__()
-
     @property
     def mandatory(self):
         return True
 
-    def first(self, root_grammar=None):
-        first = TokenLookup(self, root_grammar=root_grammar)
-        if self.parent:
-            first.add_prefix(self.position)
-        return [first]
-
 
 class InputElementHolder(InputElement):
     pass
@@ -521,30 +239,6 @@ class InputElementCollection(InputElementHolder):
         self.copy_extras(cp)
         return cp
 
-    def get(self, position):
-        return self._value[position]
-
-    def index(self, child):
-        pos = 0
-        for c in self._value:
-            if id(c) == id(child):
-                return pos
-            pos += 1
-        raise IndexError("{} doesnt have {}".format(self, child))
-
-    def first(self, root_grammar=None):
-        firsts = []
-        for i in range(len(self)):
-            elem = self.get(i)
-            firsts += elem.first(root_grammar=root_grammar)
-            if elem.mandatory:
-                break
-
-        if self.parent:
-            for first in firsts:
-                first.add_prefix(self.position)
-        return firsts
-
 
 class SequenceInputElement(InputElementCollection):
 
@@ -563,26 +257,41 @@ class SequenceInputElement(InputElementCollection):
         else:
             return len(self._value)
 
-    def get(self, position):
-        if position > 0 and self.repeat_count > 1 and position < self.repeat_count:
-            assert len(self._value) == 1
-            elem = self._value[0].copy()
-            elem.parent = self
-            elem._position = position
-            return elem
-        return self._value[position]
+    @property
+    def mandatory(self):
+        if self._mandatory is not None:
+            return self._mandatory
+        for elem in self._value:
+            if elem.mandatory:
+                self._mandatory = True
+                return True
+        self._mandatory = False
+        return False
 
-    def index(self, child):
-        if self.repeat_count > 1 and child.position < self.repeat_count:
-            if child.parent == self:
-                if child.position < len(self):
-                    return child.position
-        pos = 0
-        for c in self._value:
-            if id(c) == id(child):
-                return pos
-            pos += 1
-        raise IndexError("{} doesnt have {}".format(self, child))
+
+class OrderlessSetInputElement(InputElementCollection):
+
+    def __init__(self, sequence):
+        super().__init__(sequence)
+        self._mandatory = None
+        if isinstance(sequence, tuple):
+            self._value = sequence
+        else:
+            raise ValueError("Expected tuple object")
+
+    def __len__(self):
+        return len(self._value)
+
+    def __repr__(self, verbose=True): # noqa
+        return "Set" + '{' + ", ".join([e.__repr__(verbose=False) for e in self.value]) + '}'
+
+    def copy(self):
+        val = tuple([v.copy() for v in self._value])
+        cp = self.__class__(val)
+        for v in val:
+            v.parent = cp
+        self.copy_extras(cp)
+        return cp
 
     @property
     def mandatory(self):
@@ -620,15 +329,6 @@ class AlternativeInputElement(InputElementCollection):
         self._mandatory = True
         return True
 
-    def first(self, root_grammar=None):
-        firsts = []
-        for i in self._value:
-            firsts += i.first(root_grammar=root_grammar)
-        if self.parent:
-            for first in firsts:
-                first.add_prefix(self.position)
-        return firsts
-
 
 class NamedGrammar(InputElementHolder):
 
@@ -665,16 +365,6 @@ class NamedGrammar(InputElementHolder):
     def value(self):
         return self._value
 
-    def get(self, position):
-        if position == 0:
-            return self._value
-        raise IndexError("NamedGrammar: Invalid index: {}".format(position))
-
-    def index(self, child):
-        if id(self._value) == id(child):
-            return 0
-        raise IndexError("{} doesnt have {}".format(self, child))
-
     @property
     def parent(self):
         return None
@@ -683,10 +373,6 @@ class NamedGrammar(InputElementHolder):
     def parent(self, p):
         raise ValueError("Cannot set parent for named grammar")
 
-    @property
-    def position(self):
-        return 0
-
     @property
     def name(self):
         return self._name
@@ -695,12 +381,6 @@ class NamedGrammar(InputElementHolder):
     def mandatory(self):
         return self._value.mandatory
 
-    def first(self, root_grammar=None):
-        firsts = self._value.first(root_grammar=root_grammar)
-        for first in firsts:
-            first.add_prefix(self.position)
-        return firsts
-
 
 def map_grammar_arguments(grammar_name, parameter_list, arglist):
     arg_count = len(arglist)
@@ -742,6 +422,7 @@ def map_grammar_arguments(grammar_name, parameter_list, arglist):
 
     return param_map
 
+
 class GrammarRefElement(InputElement):
 
     def __init__(self, grammar, arglist):
@@ -775,22 +456,10 @@ class GrammarRefElement(InputElement):
     def param_mapping(self):
         return self._param_mapping
 
-    def get(self, position):
-        if position == 0:
-            return self._value
-        raise IndexError("GrammarRefElement: Invalid index: {}".format(position))
-
     @property
     def mandatory(self):
         return self._value.mandatory
 
-    def first(self, root_grammar=None):
-        firsts = self._value.first(root_grammar=root_grammar)
-        if self.parent:
-            for first in firsts:
-                first.add_prefix(self.position)
-        return firsts
-
 
 class UnresolvedInputElement(InputElement):
 
@@ -885,4 +554,338 @@ class GrammarSpecification:
     def get_tokendef(self, name):
         if name in self.token_defs:
             return self.token_defs[name]
-        return None
\ No newline at end of file
+        return None
+
+
+class LookupToken():
+
+    def __init__(self, node):
+        self._lookup_path = {}
+        self._node = node
+        self._name = self._node.element.value
+
+    @property
+    def element(self):
+        return self._node.element
+
+    @property
+    def node(self):
+        return self._node
+
+    @property
+    def name(self):
+        return self._name
+
+    @property
+    def helpstring(self):
+        return self._node.element.helpstring
+
+    @property
+    def lookup_path(self):
+        return self._lookup_path
+
+    def add_lookup_path(self, node, position):
+        if not node.path in self._lookup_path:
+            self._lookup_path[node.path] = []
+        self._lookup_path[node.path].append(position)
+
+    def copy_lookup_path_from(self, previous, node=None):
+        if node:
+            if node.path in previous.lookup_path:
+                self._lookup_path[node.path] = previous.lookup_path[node.path].copy()
+            else:
+                self._lookup_path[node.path] = []
+        else:
+            self._lookup_path = previous.lookup_path.copy()
+
+    def path_present(self, node):
+        if node.parent.path in self._lookup_path:
+            if node.position in self._lookup_path[node.parent.path]:
+                return True
+        return False
+
+    def next(self):
+        nexts = []
+        parent = self._node.parent
+        position = self._node.position + 1
+        end_of_grammar = True
+
+        while parent:
+            if not isinstance(parent.element, AlternativeInputElement):
+                while True:
+                    if isinstance(parent.element, OrderlessSetInputElement):
+                        elem = None
+                        firsts = []
+                        _options = []
+                        for i in range(parent.child_count):
+                            _elem = parent.get(i)
+                            if not self.path_present(_elem):
+                                _options.append(_elem)
+                                _firsts = _elem.first()
+                                for f in _firsts:
+                                    f.copy_lookup_path_from(self, parent)
+                                    f.add_lookup_path(parent, i)
+                                firsts += _firsts
+                        if _options and any([o.mandatory for o in _options]):
+                            end_of_grammar = False
+                    else:
+                        if position >= parent.child_count:
+                            break
+                        elem = parent.get(position)
+                        if elem.element.repeat_count > 1:
+                            firsts = elem.get(0).first()
+                            for f in firsts:
+                                f.copy_lookup_path_from(self, parent)
+                                f.add_lookup_path(parent, position)
+                        else:
+                            firsts = elem.first()
+                            for f in firsts:
+                                f.copy_lookup_path_from(self, parent)
+                                f.add_lookup_path(parent, position)
+
+                    temp_parent = parent
+                    while temp_parent.parent:
+                        for f in firsts:
+                            f.copy_lookup_path_from(self, temp_parent.parent)
+                            f.add_lookup_path(temp_parent.parent, temp_parent.position)
+                        temp_parent = temp_parent.parent
+
+                    nexts += firsts
+                    if elem and elem.mandatory:
+                        end_of_grammar = False
+
+                    if not end_of_grammar:
+                        break
+                    position += 1
+
+                    if parent.repeat_count > 1:
+                        break
+
+                    if isinstance(parent.element, OrderlessSetInputElement):
+                        break
+
+                if not end_of_grammar:
+                    break
+
+            position = parent.position + 1
+
+            parent = parent.parent
+
+        if end_of_grammar or not nexts:
+            nexts.append(EndOfInpuToken)
+
+        return nexts
+
+
+class TreeNode():
+
+    def __init__(self, tree, parent, position, element):
+        self._parent = parent
+        self._tree = tree
+        self._path = parent.path + (position,) if parent else (position, )
+        self._element = element
+        self._children = {}
+        self.reset()
+
+    def reset(self):
+        self._input_sequence = []
+        self._named_vars = {}
+        self._numbered_vars = {}
+        self._parents = []
+
+    @property
+    def child_count(self):
+        if isinstance(self._element, InputElementCollection):
+            return len(self._element.value) * self._element.repeat_count
+        else:
+            return 1
+
+    @property
+    def parents(self):
+        if not self._parents:
+            node = self
+            while node.parent:
+                self._parents.append(node)
+                node = node.parent
+        return self._parents
+
+    @property
+    def tree(self):
+        return self._tree
+
+    @property
+    def input_sequence(self):
+        return self._input_sequence
+
+    @property
+    def named_variables(self):
+        return self._named_vars
+
+    @property
+    def token_variables(self):
+        return self._numbered_vars
+
+    @property
+    def parent(self):
+        return self._parent
+
+    @property
+    def element(self):
+        return self._element
+
+    @property
+    def position(self):
+        return self._path[-1]
+
+    @property
+    def path(self):
+        return self._path
+
+    @property
+    def mandatory(self):
+        return self._element.mandatory
+
+    @property
+    def repeat_count(self):
+        return self._element.repeat_count
+
+    def add_named_variable(self, var):
+        self._named_vars[var.var_id] = var
+
+    def add_numbered_variable(self, var):
+        self._numbered_vars[var.var_id] = var
+
+    def get(self, position):
+
+        if position in self._children:
+            return self._children[position]
+
+        if position > 0 and self._element.repeat_count > 1 and position < self._element.repeat_count:
+            assert len(self._element.value) == 1
+            elem = self._element.value[0].copy()
+            elem.parent = self._element
+            elem._position = position
+            self._children[position] = TreeNode(self.tree, self, position, elem)
+        elif isinstance(self._element, InputElementCollection):
+            self._children[position] = TreeNode(self.tree, self, position, self._element.value[position])
+        elif isinstance(self._element, NamedGrammar) or isinstance(self._element, GrammarRefElement):
+            self._children = [TreeNode(self.tree, self, 0, self._element.value)]
+        else:
+            self._children = [TreeNode(self.tree, self, 0, self._element)]
+        return self._children[position]
+
+    def first(self):
+        if isinstance(self._element, InputElementCollection):
+            firsts = []
+            for i in range(self.child_count):
+                elem = self.get(i)
+                elem_firsts = elem.first()
+                for f in elem_firsts:
+                    f.add_lookup_path(self, i)
+                firsts += elem_firsts
+
+                if isinstance(self.element, AlternativeInputElement) or isinstance(self.element, OrderlessSetInputElement):
+                    continue
+
+                if elem.mandatory:
+                    break
+        elif isinstance(self._element, NamedGrammar) or isinstance(self._element, GrammarRefElement):
+            firsts = self.get(0).first()
+            for f in firsts:
+                f.add_lookup_path(self, 0)
+        else:
+            firsts = [LookupToken(self)]
+        return firsts
+
+
+
+class GrammarWalkTree(TreeNode):
+
+    def __init__(self, grammar):
+        super().__init__(self, None, 0, grammar)
+
+
+class _EndOfInpuToken(LookupToken, CliToken):
+
+    __instance = None
+
+    @property
+    def completable(self):
+        return False
+
+    def get_value(self, match_string=None, cli=None): # noqa
+        return NullTokenValue
+
+    @staticmethod
+    def getInstance():
+        if _EndOfInpuToken.__instance == None:
+            _EndOfInpuToken()
+        return _EndOfInpuToken.__instance
+
+    def __init__(self):
+        if _EndOfInpuToken.__instance != None:
+            raise Exception("This class is a singleton!")
+        self._hash = hash(_EndOfInpuToken)
+        self._name = "$END_OF_INPUT"
+        _EndOfInpuToken.__instance = self
+
+    def __repr__(self):
+        return "< End of Input >"
+
+    def __str__(self):
+        return self.__repr__()
+
+    def __not__(self):
+        return True
+
+    def truth(self):
+        return False
+
+    def __eq__(self, rhs):
+        if isinstance(rhs, _EndOfInpuToken):
+            return True
+        return False
+
+    def __ne__(self, rhs):
+        if isinstance(rhs, _EndOfInpuToken):
+            return False
+        return True
+
+    def is_(self, rhs):
+        if isinstance(rhs, _EndOfInpuToken):
+            return True
+        return False
+
+    def is_not(self, rhs):
+        if isinstance(rhs, _EndOfInpuToken):
+            return False
+        return True
+
+    def __nonzero__(self):
+        return False
+
+    def __bool__(self):
+        return False
+
+    def complete(self, token_input=None, cli=None, *args, **kwargs): # noqa
+        return []
+
+    def match(self, token_input, cli=None): # noqa
+        return MATCH_FAILURE
+
+    @property
+    def helpstring(self):
+        return str(self)
+
+    @property
+    def name(self):
+        return self._name
+
+    def next(self):
+        return [self]
+
+    def __hash__(self):
+        return self._hash
+
+
+EndOfInpuToken = _EndOfInpuToken.getInstance()
